# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Functions for working with Array dispatch."""

import functools
from types import FunctionType
from typing import Callable

from .array import Array


def wrap(
    func: Callable, wrap_return: bool = True, wrap_args: bool = True, decorator: bool = False
) -> Callable:
    """Wrap an array backend function to work with Arrays.

    Args:
        func: a function to wrap.
        wrap_return: If ``True`` convert results that are registered array
                     backend types into Array objects (Default: True).
        wrap_args: If ``True`` also wrap function type args and kwargs of the
                   wrapped function.
        decorator: If ``True`` the wrapped decorator function ``func`` will
                   also wrap the decorated functions (Default: False).

    Returns:
        Callable: The wrapped function.
    """
    if decorator:
        return _wrap_decorator(func, wrap_return=wrap_return, wrap_args=wrap_args)
    else:
        return _wrap_function(func, wrap_return=wrap_return, wrap_args=wrap_args)


def _wrap_array_function(func: Callable) -> Callable:
    """Wrap a function to handle Array-like inputs and returns"""

    @functools.wraps(func)
    def wrapped_function(*args, **kwargs):

        # Unwrap inputs
        args = tuple(
            x.__qiskit_array__().data if hasattr(x, "__qiskit_array__") else x for x in args
        )
        kwargs = dict(
            (key, val.__qiskit_array__().data) if hasattr(val, "__qiskit_array__") else (key, val)
            for key, val in kwargs.items()
        )

        # Evaluate function with unwrapped inputs
        result = func(*args, **kwargs)

        # Unwrap result
        if isinstance(result, tuple):
            result = tuple(
                x.__qiskit_array__().data if hasattr(x, "__qiskit_array__") else x for x in result
            )
        elif hasattr(result, "__qiskit_array__"):
            result = result.__qiskit_array__().data
        return result

    return wrapped_function


def _wrap_args(args):
    """Return wrapped args"""
    return tuple(_wrap_array_function(x) if isinstance(x, FunctionType) else x for x in args)


def _wrap_kwargs(kwargs):
    """Return wrapped kwargs"""
    return dict(
        (key, _wrap_array_function(val)) if isinstance(val, FunctionType) else (key, val)
        for key, val in kwargs.items()
    )


def _wrap_function(func: Callable, wrap_return: bool = True, wrap_args: bool = True) -> Callable:
    """Wrap an array backend function to work with Arrays.

    Args:
        func: a function to wrap.
        wrap_return: If ``True`` convert results that are registered array
                     backend types into Array objects (Default: True).
        wrap_args: If ``True`` also wrap function type args and kwargs of the
                   wrapped function.

    Returns:
        Callable: The wrapped function.
    """
    # pylint: disable = function-redefined
    if wrap_return and wrap_args:

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            args = _wrap_args(args)
            kwargs = _wrap_kwargs(kwargs)
            result = _wrap_array_function(func)(*args, **kwargs)
            return Array._wrap(result)

        return wrapped_func

    elif wrap_return and not wrap_args:

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            result = _wrap_array_function(func)(*args, **kwargs)
            return Array._wrap(result)

        return wrapped_func

    elif not wrap_return and wrap_args:

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            args = _wrap_args(args)
            kwargs = _wrap_kwargs(kwargs)
            return _wrap_array_function(func)(*args, **kwargs)

        return wrapped_func

    else:

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            return _wrap_array_function(func)(*args, **kwargs)

        return wrapped_func


def _wrap_decorator(func: Callable, wrap_return: bool = True, wrap_args: bool = True) -> Callable:
    """Wrap a function decorator to work with Arrays.

    Args:
        func: a function to wrap.
        wrap_return: If ``True`` convert results that are registered array
                     backend types into Array objects (Default: True).
        wrap_args: If ``True`` also wrap function type args and kwargs of the
                   wrapped function.

    Returns:
        Callable: The wrapped function.
    """
    # pylint: disable = function-redefined
    if wrap_return and wrap_args:

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            args = _wrap_args(args)
            kwargs = _wrap_kwargs(kwargs)
            decorated = _wrap_array_function(func)(*args, **kwargs)

            @functools.wraps(args[0])
            def wrapped_decorated(*f_args, **f_kwargs):
                f_args = _wrap_args(f_args)
                f_kwargs = _wrap_kwargs(f_kwargs)
                result = _wrap_function(decorated)(*f_args, **f_kwargs)
                return Array._wrap(result)

            return wrapped_decorated

        return wrapped_func

    if wrap_return and not wrap_args:

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            decorated = _wrap_array_function(func)(*args, **kwargs)

            @functools.wraps(args[0])
            def wrapped_decorated(*f_args, **f_kwargs):
                result = _wrap_function(decorated)(*f_args, **f_kwargs)
                return Array._wrap(result)

            return wrapped_decorated

        return wrapped_func

    if not wrap_return and wrap_args:

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            args = _wrap_args(args)
            kwargs = _wrap_kwargs(kwargs)
            decorated = _wrap_array_function(func)(*args, **kwargs)

            @functools.wraps(args[0])
            def wrapped_decorated(*f_args, **f_kwargs):
                f_args = _wrap_args(f_args)
                f_kwargs = _wrap_kwargs(f_kwargs)
                return _wrap_function(decorated)(*f_args, **f_kwargs)

            return wrapped_decorated

        return wrapped_func

    else:

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            decorated = _wrap_array_function(func)(*args, **kwargs)

            @functools.wraps(args[0])
            def wrapped_decorated(*f_args, **f_kwargs):
                return _wrap_function(decorated)(*f_args, **f_kwargs)

            return wrapped_decorated

        return wrapped_func



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

r"""
Compute perturbation theory terms for an LMDE.
"""

from typing import List, Optional, Callable

from scipy.integrate._ivp.ivp import OdeResult  # pylint: disable=unused-import

from multiset import Multiset

from qiskit import QiskitError

from qiskit_dynamics.array import Array
from qiskit_dynamics.perturbation.multiset_utils import _clean_multisets
from qiskit_dynamics.perturbation.perturbation_utils import (
    _merge_multiset_expansion_order_labels,
    _merge_list_expansion_order_labels,
)

from qiskit_dynamics.perturbation.dyson_magnus import (
    _solve_lmde_dyson,
    _solve_lmde_magnus,
    _solve_lmde_dyson_jax,
    _solve_lmde_magnus_jax,
)


def solve_lmde_perturbation(
    perturbations: List[Callable],
    t_span: Array,
    expansion_method: str,
    expansion_order: Optional[int] = None,
    expansion_labels: Optional[List[Multiset]] = None,
    perturbation_labels: Optional[List[Multiset]] = None,
    generator: Optional[Callable] = None,
    y0: Optional[Array] = None,
    dyson_in_frame: Optional[bool] = True,
    integration_method: Optional[str] = "DOP853",
    t_eval: Optional[Array] = None,
    **kwargs,
) -> OdeResult:
    r"""Compute time-dependent perturbation theory terms for an LMDE.

    This function computes multi-variable Dyson or Magnus expansion terms
    via the algorithm in :footcite:`puzzuoli_sensitivity_2022`, or Dyson-like terms
    via the algorithm in :footcite:`haas_engineering_2019`. See the
    :ref:`review on time-dependent perturbation theory <perturbation review>`
    to understand the details and notation used in this documentation.

    Which expansion is used is specified by the ``expansion_method`` argument, which impacts
    the interpretation of several of the function arguments (described below).
    Regardless of ``expansion_method``, the main computation is performed by
    solving a differential equation, utilizing :func:`.solve_ode`,
    and as such several of the function arguments are direct inputs into this function:

        - ``integration_method`` is the ODE method used (passed as ``method``
          to :func:`.solve_ode`), ``t_span`` is the integration interval,
          and ``t_eval`` is an optional set of points to evaluate the perturbation terms at.
        - ``kwargs`` are passed directly to :func:`.solve_ode`, enabling
          passing through of tolerance or step size arguments.

    Other arguments which are treated the same regardless off ``expansion_method`` are:

        - ``generator`` is the unperturbed generator, and the computation is performed
          in the toggling frame of this generator.

    If ``expansion_method in ['dyson', 'magnus']``, this function computes either
    multivariable Dyson series or Magnus expansion terms.
    That is, given a (finitely truncated) power series for the generator:

    .. math::

        G(t, c_0, \dots, c_{r-1}) = G_\emptyset(t)
            + \sum_{k=1}^\infty \sum_{I \in \mathcal{I}_k(r)} c_I G_I(t),

    this function computes, in the toggling frame of :math:`G_\emptyset(t)` given by ``generator``,
    either a collection of multivariable Dyson terms :math:`\mathcal{D}_I(t)` or
    multivariable Magnus terms :math:`\mathcal{O}_I(t)`, whose definitions are
    given in the :ref:`perturbation theory review <perturbation review>`.
    In this case, the arguments to the function are interpreted as follows:

        - ``perturbations`` and ``perturbation_labels`` specify the truncated generator power
          series. ``perturbations`` provides a list of python callable functions for the non-zero
          :math:`G_I(t)`, and the ``perturbation_labels`` is a list of the corresponding
          multiset labels :math:`I`, in the form of ``Multiset`` instances. If not specified,
          the labels are assumed to be
          ``[Multiset({0: 1}), ..., Multiset({len(perturbations) - 1: 1})]``.
        - ``expansion_order`` and ``expansion_labels`` specify which terms in the chosen
          expansion are to be computed. ``expansion_order`` specifies that all expansion terms up
          to a given order are to be computed, and ``expansion_labels`` specifies individual terms
          to be computed, specified as ``Multiset`` instances.
          At least one of ``expansion_order`` and ``expansion_labels`` must be specified.
          If both are specified, then all terms up to ``expansion_order`` will be computed,
          along with any additional specific terms given by ``expansion_labels``.

    Note that in the above, this function requires that the Multisets consist of
    non-negative integers. Arguments requiring lists of ``Multiset`` instances also accept
    lists of any valid format acceptable to the ``Multiset`` constructor (modulo the
    non-negative integer constraint).

    If ``expansion_method == 'dyson_like'``, the setup is different. In this case,
    for a list of matrix-valued functions :math:`G_0(t), \dots, G_{r-1}(t)`,
    this function computes integrals of the form

    .. math::
        \int_{t_0}^{t_F} dt_1 \int_{t_0}^{t_1} dt_2 \dots \int_{t_0}^{t_{k-1}}dt_k
                \tilde{G}_{i_1}(t_1) \dots \tilde{G}_{i_k}(t_k),

    for lists of integers :math:`[i_1, \dots, i_k]`, and similar to the other
    cases, :math:`\tilde{G}_j(t) = V(t)^\dagger G_j(t)V(t)`, i.e. the computation
    is performed in the toggling frame specified by ``generator``.

        - ``perturbations`` gives the list of matrix functions as callables
          :math:`G_0(t), \dots, G_{r-1}(t)`.
        - ``perturbation_labels`` is not used in this mode.
        - ``expansion_order`` specifies that all possible integrals of the above form
          should be computed up to a given order
          (i.e. integrals up to a given order with all possible orderings of the
          :math:`G_0(t), \dots, G_{r-1}(t)`).
        - ``expansion_labels`` allows for specification of specific terms to be computed.
          In this case, a term is specified by a list of ``int``\s, where the length
          of the list is the order of the integral, and the :math:`G_0(t), \dots, G_{r-1}(t)`
          appear in the integral in the order given by the list.

    Finally, additional optional arguments which can be used in the
    ``'dyson'`` and ``'dyson_like'`` cases are:

        - ``dyson_in_frame`` controls which frame the results are returned in. The default
          is ``True``, in which case the results are returned as described above. If
          ``False``, the returned results include a pre-factor of :math:`V(t)`, the solution
          of the unperturbed generator, e.g.
          :math:`V(t)\mathcal{D}_I(t)`. If ``expansion_method=='magnus'``, this argument
          has no effect on the computation.
        - ``y0`` is the initial state for the LMDE given by the unperturbed generator.
          The effect of this argument on the output is to multiply all outputs by
          ``y0`` on the right.
          If ``y0`` is supplied, the argument ``dyson_in_frame`` must be ``False``,
          as the operator :math:`V(t)` is never explicitly computed, and therefore it
          cannot be removed from the results. If ``y0`` is 1d, it will first be
          transformed into a 2d column vector to ensure consistent usage of
          matrix multiplication.

    Regardless of the value of ``expansion_method``, results are returned in an
    ``OdeResult`` instance in the same manner as :func:`.solve_ode`. The result object
    stores the results of the LMDE for ``generator`` and ``y0`` in the ``y`` attribute as in
    :func:`.solve_ode` before, and the perturbation results are in the ``perturbation_results``
    attribute storing a :class:`.PerturbationResults` object, which is a
    data container with attributes:

        - ``expansion_method``: Method as specified by the user.
        - ``expansion_labels``: Index labels for all computed perturbation terms. In the case of
          the Dyson or Magnus expansion, the labels are ``Multiset`` instances and in the
          `'dyson_like'` case are lists of ``int``\s.
        - ``expansion_terms``: A 4d array storing all computed terms. The first axis indexes
          the expansion terms in the same ordering as ``expansion_labels``,
          and the second axis indexes the perturbation terms evaluated
          at the times in ``results.t`` in the same manner as ``results.y``.

    Additionally, to retrieve the term with a given label, the :meth:`.PerturbationResults.get_term`
    method of a :class:`.PerturbationResults` instance can be called
    e.g. the results for the computation for the term with label
    ``[0, 1]`` is retrievable via ``results.perturbation_results.get_term([0, 1])``.

    Args:
        perturbations: List of matrix-valued callables.
        t_span: Integration bounds.
        expansion_method: Either ``'dyson'``, ``'magnus'``, or ``'dyson_like'``.
        expansion_order: Order of perturbation terms to compute up to. Specifying this
                         argument results in computation of all terms up to the given order.
                         Can be used in conjunction with ``expansion_labels``.
        expansion_labels: Specific perturbation terms to compute. If both ``expansion_order``
                          and ``expansion_labels`` are specified, then all terms up to
                          ``expansion_order`` are computed, along with the additional terms
                          specified in ``expansion_labels``.
        perturbation_labels: Optional description of power series terms specified by
                             ``perturbations``. To only be used with ``'dyson'``
                             and ``'magnus'`` methods.
        generator: Optional frame generator. Defaults to 0.
        y0: Optional initial state for frame generator LMDE. Defaults to the identity matrix.
        dyson_in_frame: For ``expansion_method`` ``'dyson'`` or ``'dyson_like'``,
                        whether or not to remove the frame transformation pre-factor from the
                        Dyson terms.
        integration_method: Integration method to use.
        t_eval: Points at which to evaluate the system.
        **kwargs: Additional arguments to pass to ode integration method used to compute terms.

    Returns:
        OdeResult: Results object containing standard ODE results for LMDE given by ``generator``
        and ``y0``, with additional ``perturbation_results`` attribute containing the
        requested perturbation theory terms.

    Raises:
        QiskitError: If problem with inputs, either ``expansion_method`` is unsupported,
                     or both of ``expansion_order`` and ``expansion_labels`` unspecified.

    .. footbibliography::
    """

    # validation checks
    if y0 is not None:
        if "magnus" in expansion_method:
            raise QiskitError("""Argument y0 cannot be used for expansion_method=='magnus'.""")

        if dyson_in_frame:
            raise QiskitError(
                """If expansion_method in ['dyson', 'dyson_like']
                   and y0 passed, dyson_in_frame must be False."""
            )

        # if 1d in a dyson case, turn into a column vector
        if y0.ndim == 1:
            y0 = Array([y0]).transpose()

    if perturbation_labels is not None and expansion_method == "dyson_like":
        raise QiskitError(
            """perturbation_labels argument not usable with
            expansion_method='dyson_like'."""
        )

    # clean and validate perturbation_labels, and setup expansion terms to compute
    if expansion_method in ["dyson", "magnus"]:

        if perturbation_labels is None:
            perturbation_labels = [Multiset({idx: 1}) for idx in range(len(perturbations))]
        else:
            # validate perturbation_labels
            perturbations_len = len(perturbation_labels)
            perturbation_labels = _clean_multisets(perturbation_labels)
            if len(perturbation_labels) != perturbations_len:
                raise QiskitError("perturbation_labels argument contains duplicates as multisets.")

        expansion_labels = _merge_multiset_expansion_order_labels(
            perturbation_labels=perturbation_labels,
            expansion_order=expansion_order,
            expansion_labels=expansion_labels,
        )
    elif expansion_method in ["dyson_like"]:
        expansion_labels = _merge_list_expansion_order_labels(
            perturbation_num=len(perturbations),
            expansion_order=expansion_order,
            expansion_labels=expansion_labels,
        )

    if expansion_method in ["dyson", "dyson_like"]:
        dyson_like = expansion_method == "dyson_like"
        if not Array.default_backend() == "jax":
            return _solve_lmde_dyson(
                perturbations=perturbations,
                t_span=t_span,
                dyson_terms=expansion_labels,
                perturbation_labels=perturbation_labels,
                generator=generator,
                y0=y0,
                dyson_in_frame=dyson_in_frame,
                dyson_like=dyson_like,
                integration_method=integration_method,
                t_eval=t_eval,
                **kwargs,
            )
        else:
            return _solve_lmde_dyson_jax(
                perturbations=perturbations,
                t_span=t_span,
                dyson_terms=expansion_labels,
                perturbation_labels=perturbation_labels,
                generator=generator,
                y0=y0,
                dyson_in_frame=dyson_in_frame,
                dyson_like=dyson_like,
                integration_method=integration_method,
                t_eval=t_eval,
                **kwargs,
            )
    elif expansion_method == "magnus":
        if not Array.default_backend() == "jax":
            return _solve_lmde_magnus(
                perturbations=perturbations,
                t_span=t_span,
                magnus_terms=expansion_labels,
                perturbation_labels=perturbation_labels,
                generator=generator,
                y0=y0,
                integration_method=integration_method,
                t_eval=t_eval,
                **kwargs,
            )
        else:
            return _solve_lmde_magnus_jax(
                perturbations=perturbations,
                t_span=t_span,
                magnus_terms=expansion_labels,
                perturbation_labels=perturbation_labels,
                generator=generator,
                y0=y0,
                integration_method=integration_method,
                t_eval=t_eval,
                **kwargs,
            )

    # raise error if none apply
    raise QiskitError(f"expansion_method {expansion_method} not supported.")



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Utility functions for solvers.
"""

from typing import Optional, Union, List, Tuple, Callable
import numpy as np
from scipy.integrate._ivp.ivp import OdeResult

from qiskit import QiskitError

from qiskit_dynamics.array import Array
from qiskit_dynamics.models import LindbladModel


def is_lindblad_model_vectorized(obj: any) -> bool:
    """Return True if obj is a vectorized LindbladModel."""
    return isinstance(obj, LindbladModel) and ("vectorized" in obj.evaluation_mode)


def is_lindblad_model_not_vectorized(obj: any) -> bool:
    """Return True if obj is a non-vectorized LindbladModel."""
    return isinstance(obj, LindbladModel) and ("vectorized" not in obj.evaluation_mode)


def merge_t_args(
    t_span: Union[List, Tuple, Array], t_eval: Optional[Union[List, Tuple, Array]] = None
) -> Union[List, Tuple, Array]:
    """Merge ``t_span`` and ``t_eval`` into a single array without
    duplicates. Validity of the passed ``t_span`` and ``t_eval``
    follow scipy ``solve_ivp`` validation logic:
    ``t_eval`` must be contained in ``t_span``, and be strictly
    increasing if ``t_span[1] > t_span[0]`` or strictly
    decreasing if ``t_span[1] < t_span[0]``.

    Note: this is done explicitly with ``numpy``, and hence this is
    not differentiable or compilable using jax.

    If ``t_eval is None`` returns ``t_span`` with no modification.

    Args:
        t_span: Interval to solve over.
        t_eval: Time points to include in returned results.

    Returns:
        Union[List, Tuple, Array]: Combined list of times.

    Raises:
        ValueError: If one of several validation checks fail.
    """

    if t_eval is None:
        return t_span

    t_span = Array(t_span, backend="numpy")

    t_min = np.min(t_span)
    t_max = np.max(t_span)
    t_direction = np.sign(t_span[1] - t_span[0])

    t_eval = Array(t_eval, backend="numpy")

    if t_eval.ndim > 1:
        raise ValueError("t_eval must be 1 dimensional.")

    if np.min(t_eval) < t_min or np.max(t_eval) > t_max:
        raise ValueError("t_eval entries must lie in t_span.")

    diff = np.diff(t_eval)

    if np.any(t_direction * diff <= 0.0):
        raise ValueError("t_eval must be ordered according to the direction of integration.")

    # if endpoints are not included in t_span, add them
    if t_eval[0] != t_span[0]:
        t_eval = np.append(t_span[0], t_eval)

    if t_span[1] != t_eval[-1]:
        t_eval = np.append(t_eval, t_span[1])

    return Array(t_eval, backend="numpy")


def trim_t_results(
    results: OdeResult,
    t_span: Union[List, Tuple, Array],
    t_eval: Optional[Union[List, Tuple, Array]] = None,
) -> OdeResult:
    """Trim ``OdeResult`` object based on value of ``t_span`` and ``t_eval``.

    Args:
        results: Result object, assumed to contain solution at time points
                 from the output of ``validate_and_merge_t_span_t_eval(t_span, t_eval)``.
        t_span: Interval to solve over.
        t_eval: Time points to include in returned results.

    Returns:
        OdeResult: Results with only times/solutions in ``t_eval``. If ``t_eval``
                   is ``None``, does nothing, returning solver default output.
    """

    if t_eval is None:
        return results

    t_span = Array(t_span, backend="numpy")

    # remove endpoints if not included in t_eval
    if t_eval[0] != t_span[0]:
        results.t = results.t[1:]
        results.y = Array(results.y[1:])

    if t_eval[-1] != t_span[1]:
        results.t = results.t[:-1]
        results.y = Array(results.y[:-1])

    return results


def setup_args_lists(
    args_list: List[any], args_names: List[str], args_to_list: List[Callable]
) -> Tuple[List[List[any]], bool]:
    """Transform each entry of ``args_list`` into lists of the same length.

    All elements of args_list must be either given as lists of valid specifications,
    or as a singleton of a valid specification. ``args_to_list`` contains a list of functions
    that maps its corresponding entry in args_list to a valid list of singletons, along
    with a bool indicating whether the corresponding argument was already a list of singletons.
    All args are expanded to be the same length as the longest argument. For input in args_list:
        - If len(arg) == 1, it will be repeated max_len times
        - if len(arg) == max_len, nothing is done
        - If len(arg) not in (1, max_len), an error is raised

    Args:
        args_list: List of 'inputs'.
        args_names: Names of inputs, used for error raising.
        args_to_list: As described in the main function doc. Specialized error handling for
                      the entries in args_list should be part of these functions in the
                      case that the arg is neither a valid singleton, nor a valid list
                      of singletons.

    Returns:
        Tuple: First entry is the transformed version of args_list so that all entries are the
               lists of the same length. Second entry is a boolean stating whether the inputs
               were given all as single instances (False), or one was a list (True).

    Raises:
        QiskitError: If inputs have incompatible lengths.
    """

    args_as_lists = []
    args_were_lists = False
    for arg, to_list in zip(args_list, args_to_list):
        arg_as_list, arg_was_list = to_list(arg)
        args_as_lists.append(arg_as_list)
        args_were_lists = args_were_lists or arg_was_list

    arg_lens = [len(x) for x in args_as_lists]
    max_len = max(arg_lens)
    for idx, arg_len in enumerate(arg_lens):
        if arg_len not in (1, max_len):
            max_name = args_names[arg_lens.index(max_len)]

            arg_name_sequence = ""
            for name in args_names[:-1]:
                arg_name_sequence += f"{name}, "
            arg_name_sequence += f"and {args_names[-1]}"

            raise QiskitError(
                f"""If one of {arg_name_sequence} is given as a list of valid inputs, then the
                others must specify only a single input, or a list of the same length.
                {max_name} specifies {max_len} inputs, but {args_names[idx]} is of length {arg_len},
                which is incompatible."""
            )

    args_as_lists = [
        x * max_len if arg_len == 1 else x for x, arg_len in zip(args_as_lists, arg_lens)
    ]
    return args_as_lists, args_were_lists



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Perturbative solver base class.
"""

from abc import ABC, abstractmethod
from typing import Callable, List, Union

import numpy as np
from scipy.integrate._ivp.ivp import OdeResult

from qiskit import QiskitError

from qiskit_dynamics import Signal
from qiskit_dynamics.signals import SignalList
from qiskit_dynamics.array import Array
from qiskit_dynamics.dispatch.dispatch import Dispatch

from .expansion_model import ExpansionModel
from ..solver_utils import setup_args_lists

try:
    import jax.numpy as jnp
    from jax import vmap
    from jax.lax import associative_scan
except ImportError:
    pass


class _PerturbativeSolver(ABC):
    """Abstract base class for perturbation-based solvers."""

    @abstractmethod
    def __init__(self, model: ExpansionModel):
        r"""Initialize.

        Args:
            model: ExpansionModel defining solver behaviour.
        """
        self._model = model

    @property
    def model(self) -> ExpansionModel:
        """Model object storing expansion details."""
        return self._model

    def solve(
        self,
        t0: Union[float, List[float]],
        n_steps: Union[int, List[int]],
        y0: Union[Array, List[Array]],
        signals: Union[List[Signal], List[List[Signal]]],
    ) -> Union[OdeResult, List[OdeResult]]:
        """Solve given an initial time, number of steps, signals, and initial state.

        Note that this method can be used to solve a list of simulations at once, by specifying
        one or more of the arguments ``t0``, ``n_steps``, ``y0``, or ``signals`` as a list of
        valid inputs. For this mode of operation, all of these arguments must be either lists of
        the same length, or a single valid input, which will be used repeatedly.

        Args:
            t0: Initial time.
            n_steps: Number of time steps to solve for.
            y0: Initial state at time t0.
            signals: List of signals.

        Returns:
            OdeResult: Results object, or list of results objects.

        Raises:
            QiskitError: If improperly formatted arguments.
        """

        # validate and setup list of simulations
        [t0_list, n_steps_list, y0_list, signals_list], multiple_sims = setup_args_lists(
            args_list=[t0, n_steps, y0, signals],
            args_names=["t0", "n_steps", "y0", "signals"],
            args_to_list=[
                lambda x: _scalar_to_list(x, "t0"),
                lambda x: _scalar_to_list(x, "n_steps"),
                _y0_to_list,
                _signals_to_list,
            ],
        )

        all_results = []
        for args in zip(t0_list, n_steps_list, y0_list, signals_list):
            if len(args[-1]) != len(self.model.operators):
                raise QiskitError("Signals must be the same length as the operators in the model.")
            all_results.append(
                self._solve(t0=args[0], n_steps=args[1], y0=args[2], signals=args[3])
            )

        if multiple_sims is False:
            return all_results[0]

        return all_results

    @abstractmethod
    def _solve(self, t0: float, n_steps: int, y0: Array, signals: List[Signal]) -> OdeResult:
        """Solve once for a list of signals, initial state, initial time, and number of steps.

        Args:
            t0: Initial time.
            n_steps: Number of time steps to solve for.
            y0: Initial state at time t0.
            signals: List of signals.

        Returns:
            OdeResult: Results object.
        """
        pass


def _perturbative_solve(
    single_step: Callable,
    model: "ExpansionModel",
    signals: List[Signal],
    y0: np.ndarray,
    t0: float,
    n_steps: int,
) -> np.ndarray:
    """Standard python logic version of perturbative solving routine."""
    U0 = model.rotating_frame.state_out_of_frame(t0, np.eye(model.Udt.shape[0], dtype=complex))
    Uf = model.rotating_frame.state_into_frame(
        t0 + n_steps * model.dt, np.eye(U0.shape[0], dtype=complex)
    )

    sig_cheb_coeffs = model.approximate_signals(signals, t0, n_steps)

    y = U0 @ y0
    for k in range(n_steps):
        y = single_step(sig_cheb_coeffs[:, k], y)

    return Uf @ y


def _perturbative_solve_jax(
    single_step: Callable,
    model: "ExpansionModel",
    signals: List[Signal],
    y0: np.ndarray,
    t0: float,
    n_steps: int,
) -> np.ndarray:
    """JAX version of _perturbative_solve."""
    U0 = model.rotating_frame.state_out_of_frame(
        t0, jnp.eye(model.Udt.shape[0], dtype=complex)
    ).data
    Uf = model.rotating_frame.state_into_frame(
        t0 + n_steps * model.dt, jnp.eye(U0.shape[0], dtype=complex)
    ).data

    sig_cheb_coeffs = model.approximate_signals(signals, t0, n_steps).data

    y = U0 @ y0

    step_propagators = vmap(single_step)(jnp.flip(sig_cheb_coeffs.transpose(), axis=0))
    y = associative_scan(jnp.matmul, step_propagators, axis=0)[-1] @ y

    return Uf @ y


def _scalar_to_list(x, name):
    """Check if x is a scalar or a list of scalars, and convert to a list in either case."""
    was_list = False
    x_ndim = _nested_ndim(x)
    if x_ndim > 1:
        raise QiskitError(f"{name} must be either 0d or 1d.")

    if x_ndim == 1:
        was_list = True
    else:
        x = [x]

    return x, was_list


def _y0_to_list(y0):
    """Check if y0 is a single array or list of arrays, and return as a list in either case."""
    was_list = False
    if not isinstance(y0, list):
        y0 = [y0]
    else:
        was_list = True

    return y0, was_list


def _signals_to_list(signals):
    """Check if signals is a single signal specification or a list of
    such specifications, and return as a list in either case.
    """
    was_list = False
    if signals is None:
        signals = [signals]
    elif isinstance(signals, list) and isinstance(signals[0], (list, SignalList)):
        # multiple lists
        was_list = True
    elif isinstance(signals, SignalList) or (
        isinstance(signals, list) and not isinstance(signals[0], (list, SignalList))
    ):
        # single signals list
        signals = [signals]
    else:
        raise QiskitError("Signals specified in invalid format.")

    return signals, was_list


def _nested_ndim(x):
    """Determine the 'ndim' of x, which could be composed of nested lists and array types."""
    if isinstance(x, (list, tuple)):
        return 1 + _nested_ndim(x[0])
    elif issubclass(type(x), Dispatch.REGISTERED_TYPES) or isinstance(x, Array):
        return x.ndim

    # assume scalar
    return 0



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Utility functions for working with multisets.
"""

from typing import List, Optional, Tuple
import itertools

from multiset import Multiset

from qiskit import QiskitError


def _validate_non_negative_ints(multiset: Multiset) -> bool:
    """Validate that a multiset only contains non-negative integers."""

    for elem in multiset.distinct_elements():
        if not isinstance(elem, int) or not elem >= 0:
            raise QiskitError(
                "Only Multisets whose entries are non-negative integers are accepted."
            )


def _multiset_to_sorted_list(multiset: Multiset) -> List:
    """Convert multiset to a sorted list. Assumes elements of multiset can be sorted."""

    distinct_elems = list(multiset.distinct_elements())
    distinct_elems.sort()

    sorted_list = []
    for elem in distinct_elems:
        sorted_list = sorted_list + [elem] * multiset[elem]

    return sorted_list


def _sorted_multisets(multisets: List[Multiset]) -> List[Multiset]:
    """Sort in non-decreasing order according to:

    ms1 <= ms2 if len(ms1) < len(ms2), or if
    len(ms1) == len(ms2) and if
    str(_multiset_to_sorted_list(ms1)) <= str(_multiset_to_sorted_list(ms2)).
    """

    return sorted(multisets, key=lambda x: str(len(x)) + ", " + str(_multiset_to_sorted_list(x)))


def _clean_multisets(multisets: List[Multiset]) -> List[Multiset]:
    """Given a list of multisets, remove duplicates, and sort in non-decreasing order
    according to the _sorted_multisets function.
    """

    unique_multisets = []
    for multiset in multisets:
        multiset = Multiset(multiset)

        if multiset not in unique_multisets:
            unique_multisets.append(multiset)

    return _sorted_multisets(unique_multisets)


def _submultiset_filter(
    multiset_candidates: List[Multiset], multiset_list: List[Multiset]
) -> List[Multiset]:
    """Filter the list of multiset_candidates based on whether they are a
    submultiset of an element in multiset_list.
    """

    filtered_multisets = []
    for candidate in multiset_candidates:
        for multiset in multiset_list:
            if candidate <= multiset:
                filtered_multisets.append(candidate)
                break

    return filtered_multisets


def _submultisets_and_complements(
    multiset: Multiset, submultiset_bound: Optional[int] = None
) -> Tuple[List[Multiset], List[Multiset]]:
    """Return a pair of lists giving all submultisets of size smaller than
    submultiset_bound, and corresponding complements.

    Note: Submultisets and compliments are always strict submultisets.

    Args:
        multiset: The multiset to construct submultisets from.
        submultiset_bound: Strict upper bound on submultiset to include.
                           Defaults to len(multiset).

    Returns:
        Submultisets and corresponding complements.
    """

    if submultiset_bound is None or submultiset_bound > len(multiset):
        submultiset_bound = len(multiset)

    multiset_list = _multiset_to_sorted_list(multiset)

    submultisets = []
    complements = []

    for k in range(1, submultiset_bound):
        location_subsets = itertools.combinations(range(len(multiset_list)), k)
        for location_subset in location_subsets:
            subset = []
            complement = []

            for loc, multiset_entry in enumerate(multiset_list):
                if loc in location_subset:
                    subset.append(multiset_entry)
                else:
                    complement.append(multiset_entry)

            if subset not in submultisets:
                submultisets.append(subset)
                complements.append(complement)

    # convert back to proper dict representation
    formatted_submultisets = [Multiset(submultiset) for submultiset in submultisets]
    formatted_complements = [Multiset(complement) for complement in complements]

    return formatted_submultisets, formatted_complements


def _get_all_submultisets(multisets: List[Multiset]) -> List[Multiset]:
    """Given a list of multisets, return a list of all possible submultisets
    of multisets in the list, including the original multisets.

    This returned list is sorted according to the ordering of the _sorted_multisets function.

    Args:
        multisets: List of multisets (not necessarilly correctly formatted).

    Returns:
        List: Complete list of index multisets generated by the argument.
    """

    if multisets == []:
        return []

    # clean list to unique list of properly formatted terms
    multisets = _clean_multisets(multisets)

    max_order = max(map(len, multisets))

    # partition multisets according to size
    order_dict = {k: [] for k in range(1, max_order + 1)}
    for multiset in multisets:
        order = len(multiset)
        if multiset not in order_dict[order]:
            order_dict[order].append(multiset)

    # loop through orders in reverse order, adding subterms to lower levels if necessary
    for order in range(max_order, 1, -1):

        for multiset in order_dict[order]:
            submultisets = _submultisets_and_complements(multiset, 2)[1]

            for submultiset in submultisets:
                if submultiset not in order_dict[order - 1]:
                    order_dict[order - 1].append(submultiset)

    full_list = []
    for submultisets in order_dict.values():
        full_list += submultisets

    return _sorted_multisets(full_list)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Pulse schedule to Signals converter.
"""

import warnings
from typing import Dict, List, Optional
import numpy as np

from qiskit.pulse import (
    Schedule,
    Play,
    ShiftPhase,
    SetPhase,
    ShiftFrequency,
    SetFrequency,
    Waveform,
    MeasureChannel,
    DriveChannel,
    ControlChannel,
    AcquireChannel,
)
from qiskit import QiskitError

from qiskit_dynamics.signals import DiscreteSignal


class InstructionToSignals:
    """Converts pulse instructions to Signals to be used in models.

    The :class:`InstructionsToSignals` class converts a pulse schedule to a list
    of signals that can be given to a model. This conversion is done by calling
    the :meth:`get_signals` method on a schedule. The converter applies to instances
    of :class:`Schedule`. Instances of :class:`ScheduleBlock` must first be
    converted to :class:`Schedule` using the :meth:`block_to_schedule` in
    Qiskit pulse.

    The converter can be initialized
    with the optional arguments ``carriers`` and ``channels``. These arguments
    change the returned signals of :meth:`get_signals`. When ``channels`` is given
    then only the signals specified by name in ``channels`` are returned. The
    ``carriers`` dictionary allows the user to specify the carrier frequency of
    the channels. Here, the keys are the channel name, e.g. ``d12`` for drive channel
    number 12, and the values are the corresponding frequency. If a channel is not
    present in ``carriers`` it is assumed that the carrier frequency is zero.
    """

    def __init__(
        self,
        dt: float,
        carriers: Optional[Dict[str, float]] = None,
        channels: Optional[List[str]] = None,
    ):
        """Initialize pulse schedule to signals converter.

        Args:
            dt: Length of the samples. This is required by the converter as pulse
                schedule are specified in units of dt and typically do not carry the
                value of dt with them.
            carriers: A dict of carrier frequencies. The keys are the names of the channels
                      and the values are the corresponding carrier frequency.
            channels: A list of channels that the :meth:`get_signals` method should return.
                      This argument will cause :meth:`get_signals` to return the signals in the
                      same order as the channels. Channels present in the schedule but absent
                      from channels will not be included in the returned object. If None is given
                      (the default) then all channels present in the pulse schedule are returned.
        """

        self._dt = dt
        self._channels = channels

        if isinstance(carriers, list):
            warnings.warn(
                "Passing `carriers` as a list has been deprecated and will be removed"
                "as of next release. Pass `carriers` as a dict explicitly mapping "
                "channel names to frequency values instead.",
                DeprecationWarning,
            )

        self._carriers = carriers or {}

    def get_signals(self, schedule: Schedule) -> List[DiscreteSignal]:
        """
        Args:
            schedule: The schedule to represent in terms of signals. Instances of
                      :class:`ScheduleBlock` must first be converted to :class:`Schedule`
                      using the :meth:`block_to_schedule` in Qiskit pulse.

        Returns:
            a list of piecewise constant signals.
        """

        signals, phases, frequency_shifts = {}, {}, {}

        if self._channels is not None:
            schedule = schedule.filter(channels=[self._get_channel(ch) for ch in self._channels])

        for idx, chan in enumerate(schedule.channels):
            phases[chan.name] = 0.0
            frequency_shifts[chan.name] = 0.0

            # handle deprecated list case
            carrier_freq = None
            if isinstance(self._carriers, list):
                carrier_freq = self._carriers[idx]
            else:
                carrier_freq = self._carriers.get(chan.name, 0.0)

            signals[chan.name] = DiscreteSignal(
                samples=[],
                dt=self._dt,
                name=chan.name,
                carrier_freq=carrier_freq,
            )

        for start_sample, inst in schedule.instructions:
            chan = inst.channel.name
            phi = phases[chan]
            freq = frequency_shifts[chan]

            if isinstance(inst, Play):
                start_idx = len(signals[chan].samples)

                # get the instruction samples
                inst_samples = None
                if isinstance(inst.pulse, Waveform):
                    inst_samples = inst.pulse.samples
                else:
                    inst_samples = inst.pulse.get_waveform().samples

                # build sample array to append to signal
                samples = []
                for idx, sample in enumerate(inst_samples):
                    time = self._dt * (idx + start_idx)
                    samples.append(sample * np.exp(2.0j * np.pi * freq * time + 1.0j * phi))
                signals[chan].add_samples(start_sample, samples)

            if isinstance(inst, ShiftPhase):
                phases[chan] += inst.phase

            if isinstance(inst, ShiftFrequency):
                frequency_shifts[chan] += inst.frequency

            if isinstance(inst, SetPhase):
                phases[chan] = inst.phase

            if isinstance(inst, SetFrequency):
                frequency_shifts[chan] = inst.frequency - signals[chan].carrier_freq

        # ensure all signals have the same number of samples
        max_duration = 0
        for sig in signals.values():
            max_duration = max(max_duration, sig.duration)

        for sig in signals.values():
            if sig.duration < max_duration:
                sig.add_samples(
                    start_sample=sig.duration,
                    samples=np.zeros(max_duration - sig.duration, dtype=complex),
                )

        # filter the channels
        if self._channels is None:
            return list(signals.values())

        return_signals = []
        for chan_name in self._channels:
            signal = signals.get(
                chan_name, DiscreteSignal(samples=[], dt=self._dt, name=chan_name, carrier_freq=0.0)
            )

            return_signals.append(signal)

        return return_signals

    @staticmethod
    def get_awg_signals(
        signals: List[DiscreteSignal], if_modulation: float
    ) -> List[DiscreteSignal]:
        r"""
        Create signals that correspond to the output ports of an Arbitrary Waveform Generator
        to be used with IQ mixers. For each signal in the list the number of signals is double
        to create the I and Q components. The I and Q signals represent the real and imaginary
        parts, respectively, of

        .. math::
            \Omega(t) e^{i \omega_{if} t}

        where :math:`\Omega` is the complex-valued pulse envelope and :math:`\omega_{if}` is the
        intermediate frequency.

        Args:
            signals: A list of signals for which to create I and Q.
            if_modulation: The intermediate frequency with which the AWG modulates the pulse
                           envelopes.

        Returns:
            iq signals: A list of signals which is twice as long as the input list of signals.
                For each input signal get_awg_signals returns two
        """
        new_signals = []

        for sig in signals:
            new_freq = sig.carrier_freq + if_modulation

            samples_i = sig.samples
            samples_q = np.imag(samples_i) - 1.0j * np.real(samples_i)

            sig_i = DiscreteSignal(
                sig.dt,
                samples_i,
                sig.start_time,
                new_freq,
                sig.phase,
                sig.name + "_i",
            )
            sig_q = DiscreteSignal(
                sig.dt,
                samples_q,
                sig.start_time,
                new_freq,
                sig.phase,
                sig.name + "_q",
            )

            new_signals += [sig_i, sig_q]

        return new_signals

    def _get_channel(self, channel_name: str):
        """Return the channel corresponding to the given name."""

        try:
            prefix = channel_name[0]
            index = int(channel_name[1:])

            if prefix == "d":
                return DriveChannel(index)

            if prefix == "m":
                return MeasureChannel(index)

            if prefix == "u":
                return ControlChannel(index)

            if prefix == "a":
                return AcquireChannel(index)

            raise QiskitError(
                f"Unsupported channel name {channel_name} in {self.__class__.__name__}"
            )

        except (KeyError, IndexError, ValueError) as error:
            raise QiskitError(
                f"Invalid channel name {channel_name} given to {self.__class__.__name__}."
            ) from error



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Register numpy backend for Dispatch"""

import numpy
from ..exceptions import DispatchError
from ..dispatch import Dispatch

__all__ = []


@Dispatch.register_asarray("numpy", numpy.ndarray)
def _numpy_asarray(array, dtype=None, order=None):
    """Wrapper for numpy.asarray"""
    if (
        isinstance(array, numpy.ndarray)
        and order is None
        and (dtype is None or (dtype == array.dtype and array.dtype != "O"))
    ):
        return array
    ret = numpy.asarray(array, dtype=dtype, order=order)
    if ret.dtype == "O":
        raise DispatchError("Dispatch does not support numpy object arrays.")
    return ret


@Dispatch.register_repr("numpy")
def _numpy_repr(array, prefix="", suffix=""):
    """Wrapper for showing Numpy array in custom class"""
    max_line_width = numpy.get_printoptions()["linewidth"]
    array_str = numpy.array2string(
        array,
        separator=", ",
        prefix=prefix,
        suffix="," if suffix else None,
        max_line_width=max_line_width,
    )
    sep = ""
    if len(suffix) > 1:
        last_line_width = len(array_str) - array_str.rfind("\n") + 1
        if last_line_width + len(suffix) + 1 > max_line_width:
            sep = ",\n" + " " * len(prefix)
        else:
            sep = ", "
    return prefix + array_str + sep + suffix


@Dispatch.register_array_ufunc("numpy")
def _numpy_array_ufunc(ufunc, method):
    """Trival wrapper for numpy.ufunc"""
    return getattr(ufunc, method)


@Dispatch.register_array_function("numpy")
def _numpy_array_function(func):
    """Trival wrapper for numpy array function"""
    return func



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Register jax backend for Dispatch"""
# pylint: disable=import-error


try:
    import jax
    from jax.interpreters.xla import DeviceArray
    from jax.core import Tracer
    from jax.interpreters.ad import JVPTracer
    from jax.interpreters.partial_eval import JaxprTracer

    JAX_TYPES = (DeviceArray, Tracer, JaxprTracer, JVPTracer)

    try:
        # This class is not in older versions of Jax
        from jax.interpreters.partial_eval import DynamicJaxprTracer

        JAX_TYPES += (DynamicJaxprTracer,)
    except ImportError:
        pass

    from ..dispatch import Dispatch
    import numpy as np
    from .numpy import _numpy_repr

    __all__ = []

    # Custom handling of functions not in jax.numpy
    HANDLED_FUNCTIONS = {}

    @Dispatch.register_asarray("jax", JAX_TYPES)
    def _jax_asarray(array, dtype=None, order=None):
        """Wrapper for jax.numpy.asarray"""
        if (
            isinstance(array, DeviceArray)
            and order is None
            and (dtype is None or dtype == array.dtype)
        ):
            return array
        return jax.numpy.asarray(array, dtype=dtype, order=order)

    @Dispatch.register_repr("jax")
    def _jax_repr(array, prefix="", suffix=""):
        """Wrapper for showing DeviceArray"""
        if hasattr(array, "_value"):
            return _numpy_repr(array._value, prefix=prefix, suffix=suffix)
        return prefix + repr(array) + suffix

    @Dispatch.register_array_ufunc("jax")
    def _jax_array_ufunc(ufunc, method):
        """Wrapper mapping a numpy.ufunc to jax.numpy.ufunc"""
        if method != "__call__":
            return NotImplemented
        name = ufunc.__name__
        if hasattr(jax.numpy, name):
            return getattr(jax.numpy, name)
        return NotImplemented

    @Dispatch.register_array_function("jax")
    def _jax_array_function(func):
        """Wrapper mapping a numpy function to jax.numpy function"""
        if func in HANDLED_FUNCTIONS:
            return HANDLED_FUNCTIONS[func]
        name = func.__name__
        if hasattr(jax.numpy, name):
            return getattr(jax.numpy, name)
        if hasattr(jax.numpy.linalg, name):
            return getattr(jax.numpy.linalg, name)
        return NotImplemented

    # Custom function handling

    @Dispatch.implements(np.copy, HANDLED_FUNCTIONS)
    def _copy(array, order="K"):
        return jax.numpy.array(array, copy=True, order=order)

except ModuleNotFoundError:
    pass



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Register Dispatch backends"""

from .numpy import *
from .jax import *

# Set default backend if only 1 is available
from ..dispatch import Dispatch

if len(Dispatch.REGISTERED_BACKENDS) == 1:
    Dispatch.DEFAULT_BACKEND = Dispatch.REGISTERED_BACKENDS[0]

__all__ = []



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Dispatch class"""

import functools
from types import FunctionType
from typing import Optional, Union, Tuple, Callable
from .exceptions import DispatchError


class Dispatch:
    """Dispatch Numpy ufuncs to multiple array backends."""

    # Registered backend names
    REGISTERED_BACKENDS = tuple()

    # Set version of registered backend name strings
    _REGISTERED_BACKENDS = set()

    # Registered backend array types
    REGISTERED_TYPES = tuple()

    # Set version of registered backend array types
    _REGISTERED_TYPES = {}

    # Default backend. If None the backend of the input array is used.
    DEFAULT_BACKEND = None

    # Dispatch table from backend name to asarray function
    ASARRAY_DISPATCH = {}

    # Dispatch table from backend name to repr function
    REPR_DISPATCH = {}

    # Dispatch table from backend name to array_ufunc dispatcher
    ARRAY_UFUNC_DISPATCH = {}

    # Dispatch table from backend name to array_function dispatcher
    ARRAY_FUNCTION_DISPATCH = {}

    @classmethod
    def backend(
        cls, array: any, subclass: Optional[bool] = False, fallback: Optional[str] = None
    ) -> str:
        """Return the registered backend string of a array object.

        Args:
            array: an array object.
            subclass: If True check if array type is a subclass of
                      registered types.
            fallback: Fallback backend to use if array does not match any
                      registered types.

        Returns:
            str: The array backend name if the array type is registered,
                 or `use_default=True` and the type is not registered.
            None: If `use_default=False` and the array type is not registered.

        Raises:
            DispatchError: if fallback backend is not valid.
        """
        backend = cls._REGISTERED_TYPES.get(type(array), None)
        if backend is not None:
            return backend
        if subclass:
            for key, backend in cls._REGISTERED_TYPES.items():
                if isinstance(array, key):
                    return backend
        if fallback is None or fallback in cls._REGISTERED_BACKENDS:
            return fallback
        raise DispatchError(f"fallback '{fallback}' is not a registered backend.")

    @classmethod
    def validate_backend(cls, backend: str):
        """Raise an exception if backend is not registered.

        Args:
            backend: array backend name.

        Raises:
            DispatchError: if backend is not registered.
        """
        if backend not in cls._REGISTERED_BACKENDS:
            registered = cls.REGISTERED_BACKENDS if cls.REGISTERED_BACKENDS else None
            raise DispatchError(
                f"""'{backend}' is not a registered array backends
                (registered backends: {registered})"""
            )

    @classmethod
    def array_ufunc(cls, backend: str, ufunc: callable, method: str) -> callable:
        """Return the ufunc for the specified array backend

        Args:
            backend: the array backend name.
            ufunc: the numpy ufunc.
            method: the ufunc method.

        Returns:
            callable: the ufunc for the specified array backend.
        """
        return cls.ARRAY_UFUNC_DISPATCH[backend](ufunc, method)

    @classmethod
    def repr(cls, backend: str) -> callable:
        """Return the ufunc for the specified array backend

        Args:
            backend: the array backend name.

        Returns:
            callable: the repr function for the specified array backend.
        """
        return cls.REPR_DISPATCH[backend]

    @classmethod
    def array_function(cls, backend: str, func: callable) -> callable:
        """Return the array function for the specified array backend

        Args:
            backend: the array backend name.
            func: the numpy array function.

        Returns:
            callable: the function for the specified array backend.
        """
        return cls.ARRAY_FUNCTION_DISPATCH[backend](func)

    @classmethod
    def register_types(cls, name: str, array_types: Union[any, Tuple[any]]):
        """Register an asarray array backend.

        Args:
            name: A string to identify the array backend.
            array_types: A class or list of classes to associate
                         with the array backend.
        """
        cls._REGISTERED_BACKENDS.add(name)
        cls.REGISTERED_BACKENDS = tuple(cls._REGISTERED_BACKENDS)
        if not isinstance(array_types, (list, tuple)):
            array_types = [array_types]
        for atype in array_types:
            cls._REGISTERED_TYPES[atype] = name
        cls.REGISTERED_TYPES = tuple(cls._REGISTERED_TYPES.keys())

    @classmethod
    def register_asarray(
        cls, name: str, array_types: Optional[Union[any, Tuple[any]]] = None
    ) -> callable:
        """Decorator to register an asarray function for an array backend.

        The function being wrapped must have signature
        `func(arg, dtype=None, order=None)`.

        Args:
            name: A string to identify the array backend.
            array_types: Optional, the array types to register
                         for the backend.

        Returns:
            callable: the decorated function.
        """
        if array_types:
            cls.register_types(name, array_types)

        def decorator(func):
            cls.ASARRAY_DISPATCH[name] = func
            return func

        return decorator

    @classmethod
    def register_repr(
        cls, name: str, array_types: Optional[Union[any, Tuple[any]]] = None
    ) -> callable:
        """Decorator to register an asarray function for an array backend.

        The function being wrapped must have signature
        `func(arg, dtype=None, order=None)`.

        Args:
            name: A string to identify the array backend.
            array_types: Optional, the array types to register
                         for the backend.

        Returns:
            callable: the decorated function.
        """
        if array_types:
            cls.register_types(name, array_types)

        def decorator(func):
            cls.REPR_DISPATCH[name] = func
            return func

        return decorator

    @classmethod
    def register_array_ufunc(
        cls, name: str, array_types: Optional[Union[any, Tuple[any]]] = None
    ) -> callable:
        """Decorator to register a ufunc dispatch function.

        This is used for handling of dispatch of Numpy ufuncs using
        `__array_ufunc__`. The function being wrapped must have
        signature `f(ufunc, method) -> callable(*args, **kwargs)`

        Args:
            name: A string to identify the array backend.
            array_types: Optional, the array types to register
                         for the backend.

        Returns:
            callable: the decorated function.
        """
        if array_types:
            cls.register_types(name, array_types)

        def decorator(func):
            cls.ARRAY_UFUNC_DISPATCH[name] = func
            return func

        return decorator

    @classmethod
    def register_array_function(
        cls, name: str, array_types: Optional[Union[any, Tuple[any]]] = None
    ) -> callable:
        """Decorator to register an array function dispatch function.

        This is used for handling of dispatch of Numpy functions using
        `__array_function__`. The function being wrapped must have
        signature `f(func) -> callable(*args, **kwargs)`.

        Args:
            name: A string to identify the array backend.
            array_types: Optional, the array types to register
                         for the backend.

        Returns:
            callable: the decorated function.
        """
        if array_types:
            cls.register_types(name, array_types)

        def decorator(func):
            cls.ARRAY_FUNCTION_DISPATCH[name] = func
            return func

        return decorator

    @staticmethod
    def implements(np_function, dispatch_dict):
        """Register a numpy __array_function__ implementation."""

        def decorator(func):
            dispatch_dict[np_function] = func
            return func

        return decorator


def asarray(
    array: any,
    dtype: Optional[any] = None,
    order: Optional[str] = None,
    backend: Optional[str] = None,
) -> any:
    """Convert input array to an array on the specified backend.

    This functions like `numpy.asarray` but optionally supports
    casting to other registered array backends.

    Args:
        array: An array_like input.
        dtype: Optional. The dtype of the returned array. This value
                must be supported by the specified array backend.
        order: Optional. The array order. This value must be supported
                by the specified array backend.
        backend: A registered array backend name. If None the
                    default array backend will be used.

    Returns:
        array: an array object of the form specified by the backend
                kwarg.
    """
    if backend:
        Dispatch.validate_backend(backend)
    else:
        if Dispatch.DEFAULT_BACKEND:
            backend = Dispatch.DEFAULT_BACKEND
        else:
            backend = Dispatch.backend(array, fallback="numpy")
    return Dispatch.ASARRAY_DISPATCH[backend](array, dtype=dtype, order=order)


def requires_backend(backend: str) -> Callable:
    """Return a function and class decorator for checking a backend is available.

    If the the required backend is not in the list of :func:`available_backends`
    any decorated function or method will raise an exception when called, and
    any decorated class will raise an exeption when its ``__init__`` is called.

    Args:
        backend: the backend name required by class or function.

    Returns:
        Callable: A decorator that may be used to specify that a function, class,
                  or class method requires a specific backend to be installed.
    """

    def decorator(obj):
        """Specify that the decorated object requires a specifc Array backend."""

        def check_backend(descriptor):
            if backend not in Dispatch.REGISTERED_BACKENDS:
                raise DispatchError(
                    f"Array backend '{backend}' required by {descriptor} "
                    "is not installed. Please install the optional "
                    "library '{backend}'."
                )

        # Decorate a function or method
        if isinstance(obj, FunctionType):

            @functools.wraps(obj)
            def decorated_func(*args, **kwargs):
                check_backend(f"function {obj}")
                return obj(*args, **kwargs)

            return decorated_func

        # Decorate a class
        elif isinstance(obj, type):

            obj_init = obj.__init__

            @functools.wraps(obj_init)
            def decorated_init(self, *args, **kwargs):
                check_backend(f"class {obj}")
                obj_init(self, *args, **kwargs)

            obj.__init__ = decorated_init
            return obj

        else:
            raise Exception(f"Cannot decorate object {obj} that is not a class or function.")

    return decorator



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""Utilities for type handling/conversion, primarily dealing with
reshaping arrays, and handling qiskit types that wrap arrays.
"""

from typing import Union, List
from collections.abc import Iterable
import numpy as np
from scipy.sparse import issparse, spmatrix
from scipy.sparse import kron as sparse_kron
from scipy.sparse import identity as sparse_identity
from scipy.sparse.csr import csr_matrix

from qiskit.quantum_info.operators import Operator
from qiskit_dynamics.array import Array
from qiskit_dynamics.dispatch import requires_backend

try:
    from jax.experimental import sparse as jsparse
except ImportError:
    pass


class StateTypeConverter:
    """Contains descriptions of two type specifications for DE solvers/methods,
    with functions for converting states and rhs functions between
    representations.

    A type specification is a `dict` describing a specific expected type,
    e.g. an array of a given
    shape. Currently only handled types `Array`s, specified via:
        - {'type': 'array', 'shape': tuple}

    While this class stores exact type specifications, it can be
    instantiated with a concrete type and a more general type.
    This facilitates the situation in which a solver requires a 1d `Array`,
    which is specified by the type:
        - {'type': 'array', 'ndim': 1}
    """

    def __init__(self, inner_type_spec, outer_type_spec=None, order="F"):
        """Instantiate with the inner and return types for the state.

        Args:
            inner_type_spec (dict): inner type
            outer_type_spec (dict): outer type
            order (str): order argument to be used in array reshaping.
                        Defaults to `'F'`, which corresponds to column
                        stacking convention when switching between 2d and 1d
                        arrays.
        """

        self.inner_type_spec = inner_type_spec
        self.outer_type_spec = self.inner_type_spec if outer_type_spec is None else outer_type_spec
        self.order = order

    @classmethod
    def from_instances(cls, inner_y, outer_y=None, order="F"):
        """Instantiate from concrete instances. Type of instances must
        be supported by `type_spec_from_instance`. If `outer_y is None` the
        outer type is set to the inner type.

        Args:
            inner_y (array): concrete representative of inner type
            outer_y (array): concrete representative of outer type
            order (str): order argument to be used in array reshaping.

        Returns:
            StateTypeConverter: type converter as specified by args
        """
        inner_type_spec = type_spec_from_instance(inner_y)

        outer_type_spec = None
        if outer_y is not None:
            outer_type_spec = type_spec_from_instance(outer_y)

        return cls(inner_type_spec, outer_type_spec, order)

    @classmethod
    def from_outer_instance_inner_type_spec(cls, outer_y, inner_type_spec=None, order="F"):
        """Instantiate from concrete instance of the outer type,
        and an inner type-spec. The inner type spec can be either
        be fully specified, or be more general (i.e. to
        facilitate the situation in which a solver needs a 1d array).

        Accepted general data types:
            - {'type': 'array'}
            - {'type': 'array', 'ndim': 1}

        Args:
            outer_y (array): concrete outer data type
            inner_type_spec (dict): inner, potentially general, type spec
            order (str): order argument to be used in array reshaping.

        Returns:
            StateTypeConverter: type converter as specified by args

        Raises:
            Exception: if inner_type_spec is not properly specified or is
            not a handled type
        """

        # if no inner_type_spec given just instantiate both inner
        # and outer to the outer_y
        if inner_type_spec is None:
            return cls.from_instances(outer_y, order=order)

        inner_type = inner_type_spec.get("type")
        if inner_type is None:
            raise Exception("inner_type_spec needs a 'type' key.")

        if inner_type == "array":
            outer_y_as_array = Array(outer_y)

            # if a specific shape is given attempt to instantiate from a
            # reshaped outer_y
            shape = inner_type_spec.get("shape")
            if shape is not None:
                return cls.from_instances(
                    outer_y_as_array.reshape(shape, order=order), outer_y, order=order
                )

            # handle the case that ndim == 1 is given
            ndim = inner_type_spec.get("ndim")
            if ndim == 1:
                return cls.from_instances(
                    outer_y_as_array.flatten(order=order), outer_y, order=order
                )

            # if neither shape nor ndim is given, assume it can be an array
            # of any shape
            return cls.from_instances(outer_y_as_array, outer_y, order=order)

        raise Exception("inner_type_spec not a handled type.")

    def inner_to_outer(self, y):
        """Convert a state of inner type to one of outer type."""
        return convert_state(y, self.outer_type_spec, self.order)

    def outer_to_inner(self, y):
        """Convert a state of outer type to one of inner type."""
        return convert_state(y, self.inner_type_spec, self.order)

    def rhs_outer_to_inner(self, rhs):
        """Convert an rhs function f(t, y) to work on inner type."""

        # if inner and outer type specs are the same do nothing
        if self.inner_type_spec == self.outer_type_spec:
            return rhs

        def new_rhs(t, y):
            outer_y = self.inner_to_outer(y)
            rhs_val = rhs(t, outer_y)
            return self.outer_to_inner(rhs_val)

        return new_rhs

    def generator_outer_to_inner(self, generator):
        """Convert generator from outer to inner type."""
        # if inner and outer type specs are the same do nothing
        if self.inner_type_spec == self.outer_type_spec:
            return generator

        # raise exceptions based on assumptions
        if (self.inner_type_spec["type"] != "array") or (self.outer_type_spec["type"] != "array"):
            raise Exception(
                """RHS generator transformation only
                               valid for state types Array."""
            )
        if len(self.inner_type_spec["shape"]) != 1:
            raise Exception(
                """RHS generator transformation only valid
                               if inner_type is 1d."""
            )

        if self.order == "C":
            # create identity of size the second dimension of the
            # outer type
            ident = Array(np.eye(self.outer_type_spec["shape"][1]))

            def new_generator(t):
                return Array(np.kron(generator(t), ident))

            return new_generator
        elif self.order == "F":
            # create identity of size the second dimension of the
            # outer type
            ident = Array(np.eye(self.outer_type_spec["shape"][1]))

            def new_generator(t):
                return Array(np.kron(ident, generator(t)))

            return new_generator
        else:
            raise Exception("""Unsupported order for generator conversion.""")


def convert_state(y: Array, type_spec: dict, order="F"):
    """Convert the de state y into the type specified by type_spec.
    Accepted values of type_spec are given at the beginning of the file.

    Args:
        y: the state to convert.
        type_spec (dict): the type description to convert to.
        order (str): order argument for any array reshaping function.

    Returns:
        Array: converted state.
    """

    new_y = None

    if type_spec["type"] == "array":
        # default array data type to complex
        new_y = Array(y, dtype=type_spec.get("dtype", "complex"))

        shape = type_spec.get("shape")
        if shape is not None:
            new_y = new_y.reshape(shape, order=order)

    return new_y


def type_spec_from_instance(y):
    """Determine type spec from an instance."""
    type_spec = {}
    if isinstance(y, (Array, np.ndarray)):
        type_spec["type"] = "array"
        type_spec["shape"] = y.shape

    return type_spec


def vec_commutator(
    A: Union[Array, csr_matrix, List[csr_matrix]]
) -> Union[Array, csr_matrix, List[csr_matrix]]:
    r"""Linear algebraic vectorization of the linear map X -> -i[A, X]
    in column-stacking convention. In column-stacking convention we have

    .. math::
        vec(ABC) = C^T \otimes A vec(B),

    so for the commutator we have

    .. math::
        -i[A, \cdot] = -i(A \cdot - \cdot A \mapsto id \otimes A - A^T \otimes id)

    Note: this function is also "vectorized" in the programming sense for dense arrays.

    Args:
        A: Either a 2d array representing the matrix A described above,
           a 3d array representing a list of matrices, a sparse matrix, or a
           list of sparse matrices.

    Returns:
        Array: vectorized version of the map.
    """

    if issparse(A):
        # single, sparse matrix
        sp_iden = sparse_identity(A.shape[-1], format="csr")
        return -1j * (sparse_kron(sp_iden, A) - sparse_kron(A.T, sp_iden))
    if isinstance(A, list) and issparse(A[0]):
        # taken to be 1d array of 2d sparse matrices
        sp_iden = sparse_identity(A[0].shape[-1], format="csr")
        out = [-1j * (sparse_kron(sp_iden, mat) - sparse_kron(mat.T, sp_iden)) for mat in A]
        return out

    A = to_array(A)
    iden = Array(np.eye(A.shape[-1]))
    axes = list(range(A.ndim))
    axes[-1] = axes[-2]
    axes[-2] += 1
    return -1j * (np.kron(iden, A) - np.kron(A.transpose(axes), iden))


def vec_dissipator(
    L: Union[Array, csr_matrix, List[csr_matrix]]
) -> Union[Array, csr_matrix, List[csr_matrix]]:
    r"""Linear algebraic vectorization of the linear map
    X -> L X L^\dagger - 0.5 * (L^\dagger L X + X L^\dagger L)
    in column stacking convention.

    This gives

    .. math::
        \overline{L} \otimes L - 0.5(id \otimes L^\dagger L +
            (L^\dagger L)^T \otimes id)

    Note: this function is also "vectorized" in the programming sense for dense matrices.
    """

    if issparse(L):
        sp_iden = sparse_identity(L[0].shape[-1], format="csr")
        return sparse_kron(L.conj(), L) - 0.5 * (
            sparse_kron(sp_iden, L.conj().T * L) + sparse_kron(L.T * L.conj(), sp_iden)
        )
    if isinstance(L, list) and issparse(L[0]):
        # taken to be 1d array of 2d sparse matrices
        sp_iden = sparse_identity(L[0].shape[-1], format="csr")
        out = [
            sparse_kron(mat.conj(), mat)
            - 0.5
            * (sparse_kron(sp_iden, mat.conj().T * mat) + sparse_kron(mat.T * mat.conj(), sp_iden))
            for mat in L
        ]
        return out

    iden = Array(np.eye(L.shape[-1]))
    axes = list(range(L.ndim))

    L = to_array(L)
    axes[-1] = axes[-2]
    axes[-2] += 1
    Lconj = L.conj()
    LdagL = Lconj.transpose(axes) @ L
    LdagLtrans = LdagL.transpose(axes)

    return np.kron(Lconj, iden) @ np.kron(iden, L) - 0.5 * (
        np.kron(iden, LdagL) + np.kron(LdagLtrans, iden)
    )


def isinstance_qutip_qobj(obj):
    """Check if the object is a qutip Qobj.

    Args:
        obj (any): Any object for testing.

    Returns:
        Bool: True if obj is qutip Qobj
    """
    if (
        type(obj).__name__ == "Qobj"
        and hasattr(obj, "_data")
        and type(obj._data).__name__ == "fast_csr_matrix"
    ):
        return True
    return False


# pylint: disable=too-many-return-statements
def to_array(op: Union[Operator, Array, List[Operator], List[Array], spmatrix], no_iter=False):
    """Convert an operator or list of operators to an Array.
    Args:
        op: Either an Operator to be converted to an array, a list of Operators
            to be converted to a 3d array, or an array (which simply gets
            returned)
        no_iter (Bool): Boolean determining whether to recursively unroll `Iterables`.
            If recurring, this should be True to avoid making each element of the
            input array into a separate Array.
    Returns:
        Array: Array version of input
    """
    if op is None:
        return op

    if isinstance(op, np.ndarray) and op.dtype != "O":
        if Array.default_backend() in [None, "numpy"]:
            return op
        else:
            return Array(op)

    if isinstance(op, Array):
        return op

    if issparse(op):
        return Array(op.toarray())

    if type(op).__name__ == "BCOO":
        return Array(op.todense())

    if isinstance(op, Iterable) and not no_iter:
        op = Array([to_array(sub_op, no_iter=True) for sub_op in op])
    elif isinstance(op, Iterable) and no_iter:
        return op
    else:
        op = Array(op)

    if op.backend == "numpy":
        return op.data
    else:
        return op


# pylint: disable=too-many-return-statements
def to_csr(
    op: Union[Operator, Array, List[Operator], List[Array], spmatrix], no_iter=False
) -> csr_matrix:
    """Convert an operator or list of operators to a sparse matrix.
    Args:
        op: Either an Operator to be converted to an sparse matrix, a list of Operators
            to be converted to a 3d sparse matrix, or a sparse matrix (which simply gets
            returned)
        no_iter (Bool): Boolean determining whether to recursively unroll `Iterables`.
            If recurring, this should be True to avoid making each element of the
            input into a separate csr_matrix.
    Returns:
        csr_matrix: Sparse matrix version of input
    """
    if op is None:
        return op

    if isinstance(op, csr_matrix):
        return op
    if isinstance_qutip_qobj(op):
        return op.data
    if isinstance(op, np.ndarray) and op.dtype == "O":
        op = list(op)
    if isinstance(op, (Array, np.ndarray)) and op.ndim < 3:
        return csr_matrix(op)

    if isinstance(op, Iterable) and not no_iter:
        return [to_csr(item, no_iter=True) for item in op]
    else:
        return csr_matrix(op)


@requires_backend("jax")
def to_BCOO(op: Union[Operator, Array, List[Operator], List[Array], spmatrix, "BCOO"]) -> "BCOO":
    """Convert input op or list of ops to a jax BCOO sparse array.

    Calls ``to_array`` to handle general conversion to a numpy or jax array, then
    builds the BCOO sparse array from the result.

    Args:
        op: Operator or list of operators to convert.

    Returns:
        BCOO: BCOO sparse version of the operator.
    """
    if op is None:
        return op

    if type(op).__name__ == "BCOO":
        return op

    return jsparse.BCOO.fromdense(to_array(op).data)


def to_numeric_matrix_type(
    op: Union[Operator, Array, spmatrix, List[Operator], List[Array], List[spmatrix]]
):
    """Given an operator, array, sparse matrix, or a list of operators, arrays, or sparse matrices,
    attempts to leave them in their original form, only converting the operator to an array,
    and converting lists as necessary. Summarized below:
    - operator is converted to array
    - spmatrix and Array are unchanged
    - lists of Arrays and sparse matrices are passed to their respective to_ functions
    - anything else is passed to to_array
    Args:
        op: An operator, array, sparse matrix, or list of operators, arrays or sparse matrices.
    Returns:
        Array: Array version of input
        csr_matrix: Sparse matrix version of input
    """

    if op is None:
        return op

    elif isinstance_qutip_qobj(op):
        return to_csr(op.data)

    elif isinstance(op, Array):
        return op
    elif isinstance(op, spmatrix):
        return op
    elif type(op).__name__ == "BCOO":
        return op
    elif isinstance(op, Operator):
        return to_array(op)

    elif isinstance(op, Iterable) and isinstance(op[0], spmatrix):
        return to_csr(op)

    elif isinstance(op, Iterable) and isinstance_qutip_qobj(op[0]):
        return to_csr(op)

    else:
        return to_array(op)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Wrapper for jax.experimental.ode.odeint
"""

from typing import Callable, Optional, Union, Tuple, List
import numpy as np
from scipy.integrate._ivp.ivp import OdeResult

from qiskit_dynamics.dispatch import requires_backend
from qiskit_dynamics.array import Array, wrap

from .solver_utils import merge_t_args, trim_t_results

try:
    from jax.experimental.ode import odeint as _odeint

    odeint = wrap(_odeint)
except ImportError:
    pass


@requires_backend("jax")
def jax_odeint(
    rhs: Callable,
    t_span: Array,
    y0: Array,
    t_eval: Optional[Union[Tuple, List, Array]] = None,
    **kwargs,
):
    """Routine for calling `jax.experimental.ode.odeint`

    Args:
        rhs: Callable of the form :math:`f(t, y)`
        t_span: Interval to solve over.
        y0: Initial state.
        t_eval: Optional list of time points at which to return the solution.
        **kwargs: Optional arguments to be passed to ``odeint``.

    Returns:
        OdeResult: Results object.
    """

    t_list = merge_t_args(t_span, t_eval)

    # determine direction of integration
    t_direction = np.sign(Array(t_list[-1] - t_list[0], backend="jax", dtype=complex))
    rhs = wrap(rhs)

    results = odeint(
        lambda y, t: rhs(np.real(t_direction * t), y) * t_direction,
        y0=Array(y0, dtype=complex),
        t=np.real(t_direction) * Array(t_list),
        **kwargs,
    )

    results = OdeResult(t=t_list, y=Array(results, backend="jax", dtype=complex))

    return trim_t_results(results, t_span, t_eval)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""
======================================
Models (:mod:`qiskit_dynamics.models`)
======================================

.. currentmodule:: qiskit_dynamics.models

This module contains classes for constructing the right-hand side of an ordinary differential
equations. In this package, a "model of a quantum system" means a description of a differential
equation used to model a physical quantum system, in this case either the
*Schrodinger equation*:

.. math::
    \dot{y}(t) = -i H(t)y(t),

where :math:`H(t)` is the Hamiltonian, or *Lindblad equation*:

.. math::
    \dot{\rho}(t) = -i[H(t), \rho(t)] +
                \sum_j g_j(t) \left(L_j\rho(t)L_j^\dagger -
                \frac{1}{2}\{L_j^\dagger L_j, \rho(t)\}\right),

where the second term is called the *dissipator* term. Each :math:`L_j` is a dissipation operator
*dissipator*, and :math:`[\cdot, \cdot]` and :math:`\{\cdot, \cdot\}` are, respectively, the
matrix commutator and anti-commutator.

The classes for representing the Schrodinger and Lindblad equations are,
respectively, :class:`~qiskit_dynamics.models.HamiltonianModel` and
:class:`~qiskit_dynamics.models.LindbladModel`. Model classes primarily
serve a *computational* purpose, and expose functions for evaluating
model expressions, such as :math:`t \mapsto H(t)` or :math:`t,y \mapsto -iH(t)y` in the
case of a Hamiltonian, and with similar functionality for
:class:`~qiskit_dynamics.models.LindbladModel`.

.. _Rotating frames:

Rotating frames
===============

Frame transformations are a common technique for solving time-dependent quantum differential
equations. For example, for a Hamiltonian, this corresponds to the transformation

.. math::
    H(t) \mapsto e^{iH_0t}(H(t) - H_0)e^{-iH_0t},

for a Hermitian operator :math:`H_0` called the *frame operator*.

.. note::
    The *frame operator* is commonly equivalently expressed as the corresponding anti-Hermitian
    operator under the association :math:`F = -iH_0`. This package refers to either :math:`F`
    or :math:`H_0` as the *frame operator*, with this association being understood.

Any model class can be transformed into a rotating frame by setting the
``rotating_frame`` property:

.. code-block:: python

    model.rotating_frame = frame_operator

where ``frame_operator`` is a specification of either :math:`H_0` or :math:`F = -iH_0`
(see the documentation for :class:`~qiskit_dynamics.models.RotatingFrame` for valid types
and behaviours).
Setting this property modifies the behaviour of the evaluation functions, e.g. a
:class:`~qiskit_dynamics.models.HamiltonianModel` will
compute :math:`e^{-tF}(-iH(t) - F)e^{tF}` in place of :math:`H(t)`.
:class:`~qiskit_dynamics.models.LindbladModel` has similar behaviour.

Internally, the model classes make use of the :class:`~qiskit_dynamics.models.RotatingFrame`
class, which is instantiated when the ``rotating_frame`` property is set. This class contains
helper functions for transforming various objects into and out of the rotating frame. This class
works directly with the anti-Hermitian form :math:`F = -iH_0`, however can be instantiated
with a Hermitian operator :math:`H_0` from which :math:`F` is automatically constructed.

Rotating wave approximation
===========================

The rotating wave approximation (RWA) is a transformation in which rapidly oscillating
time-dependent components, above a given cutoff frequency, are removed from a model.
This transformation is implemented in
:meth:`~qiskit_dynamics.models.rotating_wave_approximation`, see its documentation for
details.

.. _evaluation modes:

Numerical methods and evaluation modes
======================================

All model classes offer different underlying numerical implementations that a user can choose
using the ``evaluation_mode`` property. For example,
:class:`~qiskit_dynamics.models.HamiltonianModel` can internally use
either sparse or dense arrays to compute :math:`H(t)` or a product :math:`-iH(t)y`.
The default is dense arrays, and a model can be set to use sparse arrays via:

.. code-block:: python

    model.evaluation_mode = 'sparse'

See the ``evaluation_mode`` property for each model class for available modes.

.. note::

    When setting a rotating frame, models internally store their operators
    in the basis in which the frame operator is diagonal. In general, sparsity of an operator
    is not perserved by basis transformations. Hence, preserving internal sparsity with
    rotating frames requires more restrictive choice of frames. For example, diagonal frame
    operators exactly preserve sparsity.


Model classes
=============

.. autosummary::
   :toctree: ../stubs/

   HamiltonianModel
   LindbladModel
   GeneratorModel

Model transformations
=====================

.. autosummary::
   :toctree: ../stubs/

   RotatingFrame
   rotating_wave_approximation
"""

from .rotating_frame import RotatingFrame
from .generator_model import BaseGeneratorModel, GeneratorModel
from .hamiltonian_model import HamiltonianModel
from .lindblad_model import LindbladModel
from .rotating_wave_approximation import rotating_wave_approximation



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Dispatch module exceptions"""

from qiskit.exceptions import QiskitError


class DispatchError(QiskitError):
    """Class for dispatch module errors"""

    def __init__(self, *msg):
        """Set the error message."""
        super().__init__(*msg)
        self.msg = " ".join(msg)

    def __str__(self):
        """Return the message."""
        return repr(self.msg)
